@@ -9,15 +9,17 @@ __maintainer__ = "Richard Clubb"
 __email__ = "richard.clubb@embeduk.com"
 __status__ = "Development"
 
+import time
 import threading
 from pathlib import Path
 from typing import Callable
 
 from uds.config import Config
 from uds.factories import TpFactory
-from uds.uds_config_tool.UdsConfigTool import UdsTool
 from uds.uds_config_tool.IHexFunctions import ihexFile as ihexFileParser
 from uds.uds_config_tool.ISOStandard.ISOStandard import IsoDataFormatIdentifier
+from uds.uds_config_tool.UdsConfigTool import UdsTool
+
 
 ##
 # @brief a description is needed
@@ -27,13 +29,15 @@ class Uds(object):
     # @brief a constructor
     # @param [in] reqId The request ID used by the UDS connection, defaults to None if not used
     # @param [in] resId The response Id used by the UDS connection, defaults to None if not used
-    def __init__(self, odx = None, ihexFile=None, **kwargs):
+    def __init__(self, odx=None, ihexFile=None, **kwargs):
 
         self.__transportProtocol = Config.uds.transport_protocol
         self.__P2_CAN_Client = Config.uds.p2_can_client
         self.__P2_CAN_Server = Config.uds.p2_can_server
 
-        self.tp = TpFactory.select_transport_protocol(self.__transportProtocol, **kwargs)
+        self.tp = TpFactory.select_transport_protocol(
+            self.__transportProtocol, **kwargs
+        )
 
         # used as a semaphore for the tester present
         self.__transmissionActive_flag = False
@@ -45,10 +49,10 @@ class Uds(object):
         self.__ihexFile = ihexFileParser(ihexFile) if ihexFile is not None else None
         self.load_odx(odx)
 
-    def load_odx(self, odx_file: Path)-> None:
-        """Lod the given odx file and create the associated UDS 
+    def load_odx(self, odx_file: Path) -> None:
+        """Load the given odx file and create the associated UDS
         diagnostic services:
-        
+
         :param odx_file: idx file full path
         """
         if odx_file is None:
@@ -56,18 +60,18 @@ class Uds(object):
         UdsTool.create_service_containers(odx_file)
         UdsTool.bind_containers(self)
 
-    def overwrite_transmit_method(self, func : Callable):
+    def overwrite_transmit_method(self, func: Callable):
         """override transmit method from the asscociated __connection
 
-        :param func: callable use to replace the current configured 
+        :param func: callable use to replace the current configured
             transmit method
         """
         self.tp.connection.transmit = func
 
-    def overwrite_receive_method(self, func : Callable):
+    def overwrite_receive_method(self, func: Callable):
         """override the TP reception method
 
-        :param func: callable use to replace the current 
+        :param func: callable use to replace the current
             getNextBufferedMessage method
         """
         self.tp.getNextBufferedMessage = func
@@ -118,21 +122,17 @@ class Uds(object):
     def send(self, msg, responseRequired=True, functionalReq=False, tpWaitTime=0.01):
         # sets a current transmission in progress - tester present (if running) will not send if this flag is set to true
         self.__transmissionActive_flag = True
-        # print(("__transmissionActive_flag set:",self.__transmissionActive_flag))
-
-        response = None
 
         # We're moving to threaded operation, so putting a lock around the send operation.
-        self.sendLock.acquire()
-        try:
-            a = self.tp.send(msg, functionalReq, tpWaitTime)
-        finally:
-            self.sendLock.release()
+        with self.sendLock:
+            self.tp.send(msg, functionalReq, tpWaitTime)
 
         if functionalReq is True:
             responseRequired = False
 
         # Note: in automated mode (unlikely to be used any other way), there is no response from tester present, so threading is not an issue here.
+        response = None
+        
         if responseRequired:
             while True:
                 response = self.tp.recv(self.__P2_CAN_Client)
@@ -140,15 +140,11 @@ class Uds(object):
                     break
 
         # If the diagnostic session control service is supported, record the sending time for possible use by the tester present functionality (again, if present) ...
-        try:
+        if hasattr(self, "sessionSetLastSend"):
             self.sessionSetLastSend()
-        except:
-            pass  # ... if the service isn't present, just ignore
 
         # Lets go of the hold on transmissions - allows test present to resume operation (if it's running)
         self.__transmissionActive_flag = False
-        # print(("__transmissionActive_flag cleared:",self.__transmissionActive_flag))
-
         return response
 
     ##
